from __future__ import annotations

import iotools
import office


class OfficeConfig(iotools.Config):
    """A config class granting access to an os-specific appdata directory for use by this library."""

    name = office.__name__
    default = {"default_connection": "", "connections": {}}
    systemwide = False

    # noinspection PyShadowingBuiltins
    def add_office_connection(self, connection: str, id: str, secret: str, default_email: str, is_default: bool = False) -> None:
        """Add a new office connection with the given application key, secret and default email resource."""
        self.data.connections[connection] = dict(id=id, secret=secret, default_email=default_email)
        if is_default:
            self.set_default_office_connection(connection=connection)

    def set_default_office_connection(self, connection: str):
        """Set one of the existing office connections to be the default one by its connection alias."""
        if connection in self.data.connections:
            self.data.default_connection = connection
        else:
            raise ValueError(f"Connection {connection} is not one of the currently registered connections: {', '.join(self.data.connections)}. Use {type(self).__name__}.add_connection() first.")


class BlobConfig(iotools.Config):
    """A config class granting access to an os-specific appdata directory for use by this library."""

    name = "blob"
    default = {"default_connection": "", "connections": {}}

    def add_blob_connection(self, connection: str, account: str, key: str = None, is_default: bool = False) -> None:
        """Add a new blob storage connection with the given account name and key."""
        self.data.connections[connection] = dict(url=f"https://{account}.blob.core.windows.net/", key=key)
        if is_default:
            self.set_default_blob_connection(connection=connection)

    def set_default_blob_connection(self, connection: str):
        """Set one of the existing blob storage connections to be the default one by its connection alias."""
        if connection in self.data.connections:
            self.data.default_connection = connection
        else:
            raise ValueError(f"Connection {connection} is not one of the currently registered connections: {', '.join(self.data.connections)}. Use {type(self).__name__}.add_connection() first.")



from __future__ import annotations

from typing import List

import O365.address_book as address_book

from ..attribute import Attribute, NonFilterableAttribute
from ..query import Query, BulkAction, BulkActionContext
from ..outlook.message import Message, FluentMessage


class Contact(address_book.Contact):
    """A class representing a Microsoft People Contact. Contains various methods for interacting with them and their details."""

    message_constructor = Message

    def __repr__(self) -> str:
        return f"{type(self).__name__}(name={repr(self.full_name)}, email={repr(self.main_email)})"

    def __str__(self) -> str:
        return self.name

    @property
    def message(self) -> FluentMessage:
        """A property that will create a new FluentMessage with the send target set to be this contact."""
        return FluentMessage(parent=self.new_message())

    class Attributes:
        class Name(Attribute):
            name = "given_name"

        class Surname(Attribute):
            name = "surname"

        class DisplayName(Attribute):
            name = "display_name"

        class Company(Attribute):
            name = "company_name"

        class Department(Attribute):
            name = "department"

        class Created(Attribute):
            name = "created_date_time"

        class LastModified(Attribute):
            name = "last_modified_date_time"

        class HomeAddress(Attribute):
            name = "home_address"

        class JobTitle(Attribute):
            name = "job_title"

        class Manager(Attribute):
            name = "manager"

        class MiddleName(Attribute):
            name = "middle_name"

        class Mobile(Attribute):
            name = "mobile_phone1"

        class OfficeLocation(Attribute):
            name = "office_location"

        class Profession(Attribute):
            name = "profession"

        class EmailAddresses(NonFilterableAttribute):
            name = "email_addresses"


class ContactQuery(Query):
    """A class for querying the contacts within a given collection."""

    @property
    def bulk(self) -> BulkContactAction:
        """Perform a bulk action on the resultset of this query."""
        return BulkContactAction(self)

    def execute(self) -> list[Contact]:
        """Execute this query and return any contacts that match."""
        return self._container.get_contacts(limit=self._limit, query=self._query)


class BulkContactAction(BulkAction):
    """A class representing a bulk action performed on the resultset of a contact query."""

    def delete(self) -> BulkActionContext:
        """Delete all contacts that match the query this bulk action was created from."""
        return BulkActionContext(query=self._query, action=Contact.delete)



from __future__ import annotations

from typing import Any, Optional
import os
import threading
import webbrowser

from maybe import Maybe
from miscutils import Supressor
from pathmagic import Dir

with Supressor():
    from O365 import account, FileSystemTokenBackend

if True:
    from .config import OfficeConfig as Config
    from .calendar import CalendarService
    from .outlook import OutlookService
    from .people import PeopleService
    from .token import MemoryTokenBackend, BaseTokenBackend


class Office:
    """The primary class controlling access to the entirety of this library via its attributes named after Office365 services."""

    scopes = [
        "offline_access",
        "User.Read", "User.ReadBasic.All",
        "Mail.ReadWrite", "Mail.ReadWrite.Shared", "Mail.Send", "Mail.Send.Shared",
        "Contacts.ReadWrite", "Contacts.ReadWrite.Shared",
        "Calendars.ReadWrite", "Calendars.ReadWrite.Shared",
        "Files.ReadWrite", "Files.ReadWrite.All",
        "Sites.Read.All", "Sites.ReadWrite.All",
    ]

    outlook: Optional[OutlookService] = None
    people: Optional[PeopleService] = None
    calendar: Optional[CalendarService] = None

    connection: Optional[str] = None

    def __init__(self, client_id: str, client_secret: str, token_backend: BaseTokenBackend, resource: str) -> None:
        self.config, self.token, self.resource, self._auth_state = Config(), token_backend, resource, None
        self.account = Account((client_id, client_secret), main_resource=self.resource, token_backend=self.token, office=self)

        if not self.account.con.token_backend.load_token():
            self.request_token()

        self._establish_services()

    def __repr__(self) -> str:
        return f"{type(self).__name__}(account={self.resource})"

    def request_token(self) -> None:
        """ """
        import flask

        os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1"
        auth_url, callback_url = "http://localhost:5000/", "http://localhost:5000/callback/"
        app = flask.Flask("Authentication")

        @app.route('/')
        def auth_step_one():
            url, state = self.account.con.get_authorization_url(requested_scopes=Office.scopes, redirect_uri=callback_url)
            self._auth_state = state
            return flask.redirect(url)

        @app.route('/callback/')
        def auth_step_two_callback():
            self.account.con.request_token(flask.request.url, state=self._auth_state, redirect_uri=callback_url)
            flask.request.environ.get("werkzeug.server.shutdown")()
            return "Successfully authenticated. You may now close the tab."

        threading.Timer(2, lambda: webbrowser.open(auth_url)).start()

        with Supressor():
            app.run()

    def request_token_old(self) -> None:
        """ """
        auth_url, state = self.account.connection.get_authorization_url(requested_scopes=self.scopes)
        webbrowser.open(auth_url)
        self.account.connection.request_token(input("Please follow the link that will open momentarily and grant permission. Then enter the url of the inbox page you land on.\n\n"))

    def _establish_services(self) -> None:
        self.outlook, self.people, self.calendar = OutlookService(office=self), PeopleService(office=self), CalendarService(office=self)

    @classmethod
    def from_token(cls, client_id: str, client_secret: str, token: dict, resource: str = "me") -> Office:
        return cls(client_id=client_id, client_secret=client_secret, token_backend=MemoryTokenBackend(token), resource=resource)

    @classmethod
    def from_connection(cls, connection: str = None, resource: str = None) -> Office:
        config = Config()
        connection = Maybe(connection).else_(config.data.default_connection)
        token_backend = FileSystemTokenBackend(token_path=str(config.folder.new_dir("tokens")), token_filename=f"{connection}.txt")

        settings = config.data.connections[connection]
        resource = Maybe(resource).else_(settings.default_email)
        return cls(client_id=settings.id, client_secret=settings.secret, token_backend=token_backend, resource=resource)


class Account(account.Account):
    def __init__(self, *args: Any, office: Office = None, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.con.office = office



# import pytest


class TestCalendarService:
    def test___getitem__(self):  # synced
        assert True

    def test_default(self):  # synced
        assert True

    def test_custom(self):  # synced
        assert True


class TestSchedule:
    pass



from __future__ import annotations

import O365.calendar as calendar

from .event import Event, EventQuery, FluentEvent


class Calendar(calendar.Calendar):
    """A class representing a Microsoft Outlook calendar. Provides methods and properties for interacting with it."""

    event_constructor = Event

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join([f'{attr}={repr(val)}' for attr, val in self.__dict__.items() if not attr.startswith('_')])})"

    @property
    def events(self) -> EventQuery:
        return EventQuery(container=self)

    @property
    def event(self) -> FluentEvent:
        """A property that will create a new fluent event."""
        return self.new_event().fluent

    def new_event(self, subject: str = None) -> Event:
        return super().new_event(subject=subject)



from __future__ import annotations

from typing import List

import O365.mailbox as mailbox

from .message import Message, MessageQuery
from ..attribute import Attribute, NonFilterableAttribute
from ..query import Query, BulkAction, BulkActionContext


class MessageFolder(mailbox.Folder):
    """A class representing a Microsoft Outlook message folder. Can initiate queries on any messages or folders it contains."""

    message_constructor = Message

    def __repr__(self) -> str:
        return f"{type(self).__name__}(name={self.name})"

    @property
    def folders(self) -> MessageFolderQuery:
        """A property that will create a new query against the folders contained within this folder."""
        return MessageFolderQuery(container=self)

    @property
    def messages(self) -> MessageQuery:
        """A property that will create a new query against the messages contained within this folder."""
        return MessageQuery(container=self)

    @staticmethod
    def order_messages_by_date(messages: list, descending: bool = True) -> list:
        """Order a collection of messages by datetime received."""
        return sorted(messages, reverse=descending, key=lambda val: val.received)

    class Attributes:
        class ChildFolderCount(Attribute):
            name = "child_folder_count"

        class TotalItemCount(Attribute):
            name = "total_item_count"

        class UnreadItemCount(Attribute):
            name = "unread_item_count"

        class Name(Attribute):
            name = "display_name"

        class ChildFolders(NonFilterableAttribute):
            name = "child_folders"

        class Messages(NonFilterableAttribute):
            name = "messages"


class BulkMessageFolderAction(BulkAction):
    """A class representing a bulk action performed on the resultset of a message folder query."""

    def move(self, folder: MessageFolder) -> BulkActionContext:
        """Move all folders that match the query this bulk action was created from into the given folder."""
        return BulkActionContext(query=self._query, action=MessageFolder.move_folder, args=(folder,))

    def delete(self) -> BulkActionContext:
        """Delete all folders that match the query this bulk action was created from."""
        return BulkActionContext(query=self._query, action=MessageFolder.delete)

    def copy(self, folder: MessageFolder) -> BulkActionContext:
        """Copy all folders that match the query this bulk action was created from into the given folder."""
        return BulkActionContext(query=self._query, action=MessageFolder.copy_folder, args=(folder,))


class MessageFolderQuery(Query):
    """A class for querying the message folders within a given collection."""

    def __getitem__(self, key: str) -> MessageFolder:
        return self._container.get_folder(folder_name=key)

    def execute(self) -> list[Message]:
        """Execute this query and return any folders that match."""
        return list(self._container.get_folders(limit=self._limit, query=self._query))

    @property
    def bulk(self) -> BulkMessageFolderAction:
        """Perform a bulk action on the resultset of this query."""
        return BulkMessageFolderAction(self)



# import pytest


class TestQuery:
    def test___call__(self):  # synced
        assert True

    def test_bulk(self):  # synced
        assert True

    def test_select(self):  # synced
        assert True

    def test_where(self):  # synced
        assert True

    def test_order_by(self):  # synced
        assert True

    def test_limit(self):  # synced
        assert True

    def test_execute(self):  # synced
        assert True

    def test__build_select_clause(self):  # synced
        assert True

    def test__build_where_clause(self):  # synced
        assert True

    def test__build_order_by_clause(self):  # synced
        assert True

    def test__build_boolean_expression_clause(self):  # synced
        assert True

    def test__build_side(self):  # synced
        assert True

    def test__build_chain_operator(self):  # synced
        assert True

    def test__build_boolean_expression(self):  # synced
        assert True

    def test__negation(self):  # synced
        assert True

    def test__precedence_grouping(self):  # synced
        assert True


class TestBulkActionContext:
    def test___len__(self):  # synced
        assert True

    def test___bool__(self):  # synced
        assert True

    def test_commit(self):  # synced
        assert True

    def test_execute(self):  # synced
        assert True

    def test__execute_query(self):  # synced
        assert True

    def test__perform_bulk_action(self):  # synced
        assert True


class TestBulkAction:
    pass



__all__ = ["Office", "CalendarService", "Calendar", "Event", "OutlookService", "MessageFolder", "Message", "PeopleService", "ContactFolder", "Contact", "OfficeConfig", "BlobConfig", "BlobStorage"]

from .office import Office
from .calendar import CalendarService, Calendar, Event
from .outlook import OutlookService, MessageFolder, Message
from .people import PeopleService, ContactFolder, Contact
from .config import OfficeConfig, BlobConfig
from .blob import BlobStorage



# import pytest


class TestContactFolder:
    def test_folders(self):  # synced
        assert True

    def test_contacts(self):  # synced
        assert True

    def test_from_address(self):  # synced
        assert True

    class TestAttributes:
        class TestName:
            pass

        class TestContacts:
            pass

        class TestChildFolders:
            pass


class TestBulkContactFolderAction:
    def test_move(self):  # synced
        assert True

    def test_delete(self):  # synced
        assert True


class TestContactFolderQuery:
    def test___getitem__(self):  # synced
        assert True

    def test_execute(self):  # synced
        assert True

    def test_bulk(self):  # synced
        assert True



# import pytest


class TestFluentEntity:
    def test_subject(self):  # synced
        assert True

    def test_body(self):  # synced
        assert True

    def test_attach(self):  # synced
        assert True

    def test_sign(self):  # synced
        assert True

    def test_importance(self):  # synced
        assert True

    def test_categories(self):  # synced
        assert True

    def test__parse_contacts_to_emails(self):  # synced
        assert True



from __future__ import annotations

from io import BytesIO
from typing import Iterator, TYPE_CHECKING

from subtypes import NameSpace, Str, Dict
from pathmagic import File, Dir, PathLike
from miscutils import cached_property, ReprMixin

from .config import BlobConfig as Config

if TYPE_CHECKING:
    from azure.storage.blob import ContainerClient, BlobClient


class BlobStorage:
    """A class representing a blob storage account. Takes a connection alias which must exist in the library config settings. If none is provided, the default connection will be used."""

    def __init__(self, url: str, key: str) -> None:
        from azure.storage.blob import BlobServiceClient

        self.config = Config()
        self.client = BlobServiceClient(account_url=url, credential=key)
        self.containers = BlobContainerNameSpace(self)

    def new_container(self, name: str) -> BlobContainer:
        self.client.create_container(name)
        self.containers()
        return self.containers[name]

    def blob_from_url(self, url: str) -> Blob:
        container, blob = Str(url).slice.after(f"{self.client.primary_hostname}/").split("/", 1)
        return self.containers[container][blob]

    @classmethod
    def from_connection(cls, connection: str = None) -> BlobStorage:
        config = Config()
        connection = connection or config.data.default_connection
        credentials = config.data.connections[connection]
        return cls(url=credentials.url, key=credentials.key)


class BlobContainerNameSpace(NameSpace):
    """A namespace class representing a collection of blob containers."""

    def __init__(self, storage: BlobStorage) -> None:
        self._storage = storage
        self()

    def __call__(self, *args, **kwargs) -> BlobContainerNameSpace:
        super().__call__({container.name: BlobContainer(name=container.name, storage=self._storage) for container in self._storage.client.list_containers()})
        return self

    def __getitem__(self, name: str) -> BlobContainer:
        return super().__getitem__(name=name)


class BlobContainer:
    """A class representing a blob container, with methods for accessing (supports item access) and iterating over its blobs."""

    def __init__(self, name: str, storage: BlobStorage) -> None:
        self.name, self.storage = name, storage
        self._cached_len = 0

    def __repr__(self) -> str:
        return f"{type(self).__name__}(name={repr(self.name)}, num_blobs={repr(self._cached_len or '?')})"

    def __str__(self) -> str:
        return self.name

    def __len__(self) -> int:
        self._cached_len = sum([1 for blob in self.client.list_blobs(self.name)])
        return self._cached_len

    def __bool__(self) -> bool:
        return len(self) > 0

    def __iter__(self) -> Iterator[Blob]:
        return iter(Blob(name=blob.name, container=self) for blob in self.client.list_blobs())

    def __getitem__(self, key: str) -> Blob:
        return Blob(name=key, container=self)

    @cached_property
    def client(self) -> ContainerClient:
        return self.storage.client.get_container_client(self.name)

    @cached_property
    def properties(self) -> Dict:
        return Dict(self.client.get_container_properties())

    def upload_file(self, file: PathLike, name: str = None) -> UploadAccessor:
        return self[name if name else File.from_pathlike(file).name].upload.from_file(file)

    def delete(self) -> None:
        if list(self):
            raise PermissionError(f"May not delete non-empty ({len(self)} blobs found) container '{self.name}'.")
        else:
            self.storage.client.delete_container(self.name)
            self.storage.containers()


class Blob:
    """A class representing a blob in storage within a container, with methods for interacting with it."""

    def __init__(self, name, container: BlobContainer) -> None:
        self.name, self.container = name, container

    def __repr__(self) -> str:
        return f"{type(self).__name__}(name={repr(self.name)}, container={repr(self.container.name)}, exists={bool(self)})"

    def __bool__(self) -> bool:
        return self.exists()

    @cached_property
    def client(self) -> BlobClient:
        return self.container.client.get_blob_client(self.name)

    @cached_property
    def properties(self) -> Dict:
        return Dict(self.client.get_blob_properties())

    @property
    def url(self) -> str:
        return self.client.url

    @cached_property
    def upload(self) -> UploadAccessor:
        return UploadAccessor(parent=self)

    @cached_property
    def download(self) -> DownloadAccessor:
        return DownloadAccessor(parent=self)

    def delete(self) -> None:
        """Permanently delete this blob within its container in storage."""
        self.client.delete_blob()

    def exists(self) -> bool:
        try:
            if self.properties:
                return True
        except Exception:
            return False


class UploadAccessor(ReprMixin):
    def __init__(self, parent: Blob) -> None:
        self.parent = parent
        self()

    def __call__(self, overwrite: bool = False) -> UploadAccessor:
        self.overwrite = overwrite
        return self

    def from_file(self, file: PathLike) -> Blob:
        """Create a new blob within this container in storage from the given file path."""
        file = File.from_pathlike(file)

        with open(file, "rb") as stream:
            self.parent.client.upload_blob(data=stream, overwrite=self.overwrite)

        return self.parent

    def from_bytes(self, data: bytes) -> Blob:
        """Create a new blob within this container in storage from the given file path."""
        self.parent.client.upload_blob(data=data, overwrite=self.overwrite)
        return self.parent

    def from_stream(self, stream: BytesIO) -> Blob:
        """Create a new blob within this container in storage from the given file path."""
        stream.seek(0)
        self.parent.client.upload_blob(data=stream, overwrite=self.overwrite)
        return self.parent


class DownloadAccessor(ReprMixin):
    def __init__(self, parent: Blob) -> None:
        self.parent = parent
        self()

    def __call__(self) -> UploadAccessor:
        return self

    def to_folder(self, folder: PathLike, name: str = None) -> File:
        """Download this blob to the given folder. It will keep its blob 'basename' as its new name."""
        file = Dir.from_pathlike(folder).new_file(Str(self.parent.name).slice.after_last("/") or self.parent.name if name is None else name)

        with open(file, "wb") as stream:
            self.parent.client.download_blob().readinto(stream)

        return file

    def as_file(self, file: PathLike) -> File:
        """Download this blob to the given path."""
        file = File.from_pathlike(file)

        with open(file, "wb") as stream:
            self.parent.client.download_blob().readinto(stream)

        return file

    def as_bytes(self) -> bytes:
        return self.parent.client.download_blob().readall()

    def as_stream(self, stream: BytesIO = None) -> BytesIO:
        stream = stream or BytesIO()
        self.parent.client.download_blob().readinto(stream)
        stream.seek(0)
        return stream



# import pytest


class TestOutlookService:
    def test___getitem__(self):  # synced
        assert True

    def test_message(self):  # synced
        assert True

    def test_signature(self):  # synced
        assert True

    def test_main(self):  # synced
        assert True

    def test_inbox(self):  # synced
        assert True

    def test_outbox(self):  # synced
        assert True

    def test_sent(self):  # synced
        assert True

    def test_drafts(self):  # synced
        assert True

    def test_junk(self):  # synced
        assert True

    def test_deleted(self):  # synced
        assert True

    def test_custom(self):  # synced
        assert True


class TestMailbox:
    pass



from __future__ import annotations

from typing import TYPE_CHECKING

from O365 import calendar

from miscutils import cached_property

from .calendar import Calendar
from .event import Event

if TYPE_CHECKING:
    from ..office import Office


class CalendarService:
    """A class representing Microsoft Calendar. Controls access to calendar-related services."""

    def __init__(self, office: Office) -> None:
        self.office = office
        self.schedule = Schedule(parent=self.office.account)

    def __getitem__(self, key: str) -> Calendar:
        return self.custom(calendar_name=key)

    @cached_property
    def default(self) -> Calendar:
        """A property that returns the default calendar."""
        return self.schedule.get_default_calendar()

    def custom(self, calendar_name: str = None, calendar_id: str = None) -> Calendar:
        """Return the given custom folder by name or id."""
        return self.schedule.get_calendar(calendar_name=calendar_name, calendar_id=calendar_id)


class Schedule(calendar.Schedule):
    calendar_construcor = Calendar
    event_constructor = Event



from __future__ import annotations

from typing import Any, List, Dict, TYPE_CHECKING

from O365.directory import Directory, User

from subtypes import Str, NameSpace
from miscutils import cached_property, is_running_in_ipython

from .contact import Contact
from .folder import ContactFolder

if TYPE_CHECKING:
    from ..office import Office


class PeopleService:
    """A class representing Microsoft People. Controls access to contact-related services."""

    def __init__(self, *args: Any, office: Office, **kwargs: Any) -> None:
        self.office = office

        if is_running_in_ipython():
            assert self.contacts

    @cached_property
    def contacts(self) -> ContactNameSpace:
        return ContactNameSpace(service=self)

    @cached_property
    def personal(self) -> ContactFolder:
        return ContactFolder(parent=self.office.account, main_resource=self.office.account.main_resource, name="Personal Address Book", root=True)

    @cached_property
    def active_directory(self) -> Directory:
        """A property that returns the Azure Active Directory."""
        return Directory(parent=self.office.account, main_resource="users")

    @cached_property
    def me(self) -> User:
        """A property that returns the Azure Active Directory."""
        return Directory(parent=self.office.account, main_resource="me").get_current_user()


class ContactNameSpace(NameSpace):
    """A namespace class containing a collection of the contacts within the global address book of the email address used to instanciate the Office object."""

    def __init__(self, service: PeopleService) -> None:
        contacts_by_name: dict[str, list[Contact]] = {}
        for contact in service.personal.get_contacts():
            contacts_by_name.setdefault(Str(contact.name).case.snake(), []).append(contact)
            contacts_by_name.setdefault(Str(contact.display_name).case.snake(), []).append(contact)

        mappings = {}
        for name, contacts in contacts_by_name.items():
            if len(contacts) == 1 and name and name.lower() != "none" and not hasattr(self, name):
                mappings[name] = contacts[0]

        super().__init__(mappings)



from __future__ import annotations

import contextlib
from typing import Any, Callable, Collection, Generator, Tuple, Union, Optional

import O365.utils.utils as utils

from maybe import Maybe
from miscutils import issubclass_safe

from .attribute import BaseAttribute, Attribute, BooleanAttributeMeta, FilterableAttribute, BooleanExpression, BooleanExpressionClause


class Query:
    """A class for querying the api elements within a given collection."""

    def __init__(self, container: Any) -> None:
        self._container = container
        self._casing_function = self._container.protocol.casing_function
        self._query = utils.Query(protocol=self._container.protocol)
        self._select: Optional[Tuple[BaseAttribute, ...]] = None
        self._where: Optional[BooleanExpressionClause] = None
        self._order: Optional[FilterableAttribute] = None
        self._limit: Optional[int] = None

    def __repr__(self) -> str:
        return repr(self._query)

    def __call__(self) -> Any:
        return self.execute()

    @property
    def bulk(self) -> BulkAction:
        """Perform a bulk action on the resultset of this query."""
        return BulkAction(self)

    def select(self, *args: Tuple[BaseAttribute, ...]) -> Query:
        """Set the attributes that will be queried. If this method is not called, all message attributes will be returned."""
        self._query._selects = set()
        self._select = args
        self._build_select_clause()
        return self

    def where(self, resolvable_element: Union[Attribute, BooleanExpression, BooleanExpressionClause]) -> Query:
        """Set the filter clause on this query. Accepts a single boolean attribute, boolean expression or boolean expression clause."""
        self._query.clear_filters()
        self._where = resolvable_element._resolve()
        self._build_where_clause()
        return self

    def order_by(self, order_clause: Any) -> Query:
        """Set the filter clause on this query. Accepts a single boolean attribute, boolean expression or boolean expression clause."""
        self._query.clear_order()
        self._order = order_clause
        self._build_order_by_clause()
        return self

    def limit(self, limit: int = 25) -> Query:
        """Set the limit on the number of objects that may be returned."""
        self._limit = limit
        return self

    def execute(self) -> Any:
        """Execute this query and return the results."""
        raise NotImplementedError

    def _build_select_clause(self) -> None:
        if self._select:
            self._query.select(*[self._casing_function(attribute.name) for attribute in self._select])

    def _build_where_clause(self) -> None:
        if self._where is not None:
            if isinstance(self._where, BooleanExpression):
                self._build_boolean_expression(self._where)
            elif isinstance(self._where, BooleanExpressionClause):
                self._build_boolean_expression_clause(self._where)
            else:
                raise TypeError(f"Argument to filter clause of '{type(self).__name__}' must be '{type(BooleanExpression.__name__)}' or '{type(BooleanExpressionClause.__name__)}', not '{type(self._where).__name__}'.")

    def _build_order_by_clause(self) -> None:
        if isinstance(self._order, str):
            self._query.order_by(self._order)
        elif isinstance(self._order, FilterableAttribute):
            self._query.order_by(self._casing_function(self._order.name), ascending=self._order.ascending)
        elif issubclass_safe(self._order, FilterableAttribute):
            attribute = self._order.asc()
            self._query.order_by(self._casing_function(attribute.name), ascending=self._order.ascending)
        else:
            raise TypeError(f"Unrecognized type '{type(self._order)}' of '{self._order}' for 'order_by'.")

    def _build_boolean_expression_clause(self, clause: BooleanExpressionClause) -> None:
        with self._precedence_grouping():
            self._build_side(clause.left)
            self._build_chain_operator(clause.operator)
            self._build_side(clause.right)

    def _build_side(self, side: Union[BooleanExpression, BooleanExpressionClause]) -> None:
        if isinstance(side, BooleanAttributeMeta):
            side = side._resolve()

        if isinstance(side, BooleanExpression):
            self._build_boolean_expression(side)
        elif isinstance(side, BooleanExpressionClause):
            self._build_boolean_expression_clause(side)
        else:
            raise TypeError(f"The sides of '{type(BooleanExpressionClause).__name__}' must be '{type(BooleanExpression).__name__}' or '{type(BooleanExpressionClause).__name__}', not '{type(side).__name__}'.")

    def _build_chain_operator(self, operator: utils.ChainOperator) -> None:
        self._query.chain(operator.value)

    def _build_boolean_expression(self, expression: BooleanExpression) -> None:
        self._query.on_attribute(self._casing_function(expression.attr))
        with self._negation() if expression.negated else contextlib.nullcontext():
            expression.func(self._query, expression.arg)

    @contextlib.contextmanager
    def _negation(self) -> Generator[None, None, None]:
        self._query.negate()
        yield
        self._query.negate()

    @contextlib.contextmanager
    def _precedence_grouping(self) -> Generator[None, None, None]:
        self._query.open_group()
        yield
        self._query.close_group()


class BulkActionContext:
    """A class representing the context within which a bulk action is performed. It can be used as a context manager and will automatically perform the action upon dropping out of scope if the action was committed."""

    def __init__(self, query: Query, action: Callable, args: Any = None, kwargs: Any = None) -> None:
        self._query, self._action, self._args, self._kwargs, self._committed = query, action, Maybe(args).else_(()), Maybe(kwargs).else_({}), False
        self.result: Collection = []

    def __len__(self) -> int:
        return len(self.result)

    def __bool__(self) -> bool:
        return len(self) > 0

    def __enter__(self) -> BulkActionContext:
        self._execute_query()
        return self

    def __exit__(self, ex_type: Any, ex_value: Any, ex_traceback: Any) -> None:
        if self._committed:
            self._perform_bulk_action()

    def commit(self) -> None:
        """Commit the action corresponding to this context. It will be performed when this object drops out of context."""
        self._committed = True

    def execute(self) -> int:
        """Perform the bulk action corresponding to this context."""
        self._execute_query()
        self._perform_bulk_action()
        return len(self)

    def _execute_query(self) -> None:
        self.result = self._query.execute()

    def _perform_bulk_action(self) -> None:
        for msg in self.result:
            self._action(msg, *self._args, **self._kwargs)


class BulkAction:
    """A class representing a bulk action performed on the resultset of a query."""

    def __init__(self, query: Query) -> None:
        self._query = query



# import pytest


class TestMessageFolder:
    def test_folders(self):  # synced
        assert True

    def test_messages(self):  # synced
        assert True

    def test_order_messages_by_date():  # synced
        assert True

    class TestAttributes:
        class TestChildFolderCount:
            pass

        class TestTotalItemCount:
            pass

        class TestUnreadItemCount:
            pass

        class TestName:
            pass

        class TestChildFolders:
            pass

        class TestMessages:
            pass


class TestBulkMessageFolderAction:
    def test_move(self):  # synced
        assert True

    def test_delete(self):  # synced
        assert True

    def test_copy(self):  # synced
        assert True


class TestMessageFolderQuery:
    def test___getitem__(self):  # synced
        assert True

    def test_execute(self):  # synced
        assert True

    def test_bulk(self):  # synced
        assert True



# import pytest


class TestCalendar:
    def test_events(self):  # synced
        assert True

    def test_event(self):  # synced
        assert True

    def test_new_event(self):  # synced
        assert True



# import pytest


class TestOffice:
    def test_request_token(self):  # synced
        assert True

    def test__establish_services(self):  # synced
        assert True


class TestAccount:
    pass



# import pytest


class TestBlobStorage:
    def test_new_container(self):  # synced
        assert True


class TestBlobContainerNameSpace:
    def test___call__(self):  # synced
        assert True


class TestBlobContainer:
    def test___str__(self):  # synced
        assert True

    def test___len__(self):  # synced
        assert True

    def test___bool__(self):  # synced
        assert True

    def test___iter__(self):  # synced
        assert True

    def test___getitem__(self):  # synced
        assert True

    def test_client(self):  # synced
        assert True

    def test_download_blob_to(self):  # synced
        assert True

    def test_download_blob_as(self):  # synced
        assert True

    def test_upload_blob_from(self):  # synced
        assert True

    def test_delete(self):  # synced
        assert True


class TestBlob:
    def test_client(self):  # synced
        assert True

    def test_download_to(self):  # synced
        assert True

    def test_download_as(self):  # synced
        assert True

    def test_delete(self):  # synced
        assert True



# import pytest


class TestMessage:
    def test___str__(self):  # synced
        assert True

    def test___hash__(self):  # synced
        assert True

    def test__repr_html_(self):  # synced
        assert True

    def test_text(self):  # synced
        assert True

    def test_markup(self):  # synced
        assert True

    def test_fluent(self):  # synced
        assert True

    def test_reply(self):  # synced
        assert True

    def test_forward(self):  # synced
        assert True

    def test_copy(self):  # synced
        assert True

    def test_render(self):  # synced
        assert True

    def test_save_attachments_to(self):  # synced
        assert True

    class TestAttributes:
        class TestFrom:
            pass

        class TestSender:
            pass

        class TestSubject:
            pass

        class TestReceivedOn:
            pass

        class TestLastModified:
            pass

        class TestCategories:
            pass

        class TestIsRead:
            pass

        class TestHasAttachments:
            pass

        class TestIsDraft:
            pass

        class TestHasDeliveryReceipt:
            pass

        class TestHasReadReceipt:
            pass

        class TestImportance:
            pass

        class TestBody:
            pass

        class TestCc:
            pass

        class TestBcc:
            pass

        class TestTo:
            pass


class TestBulkMessageAction:
    def test_copy(self):  # synced
        assert True

    def test_move(self):  # synced
        assert True

    def test_delete(self):  # synced
        assert True

    def test_mark_as_read(self):  # synced
        assert True

    def test_save_draft(self):  # synced
        assert True


class TestMessageQuery:
    def test_bulk(self):  # synced
        assert True

    def test_execute(self):  # synced
        assert True


class TestFluentMessage:
    def test_from_(self):  # synced
        assert True

    def test_to(self):  # synced
        assert True

    def test_cc(self):  # synced
        assert True

    def test_bcc(self):  # synced
        assert True

    def test_request_read_receipt(self):  # synced
        assert True

    def test_request_delivery_receipt(self):  # synced
        assert True

    def test_send(self):  # synced
        assert True



from __future__ import annotations

from typing import Union, List, Collection, TYPE_CHECKING
from collections.abc import Iterable
import os

from O365.utils import ImportanceLevel

from pathmagic import PathLike

if TYPE_CHECKING:
    from .outlook.message import Message
    from .people.contact import Contact
    from .calendar.event import Event


class FluentEntity:
    """A class representing an entity that doesn't yet exist. All public methods allow chaining."""
    entity: Union[Message, Event] = None
    _signing: bool
    _temp_body: str

    def subject(self, subject: str) -> FluentEntity:
        """Set the subject of the message."""
        self.entity.subject = subject
        return self

    def body(self, body: str) -> FluentEntity:
        """Set the body of the message. The body should be an html string, but python newline and tab characters will be automatically converted to their html equivalents."""
        self._temp_body = body.replace("\n", "<br>").replace("\t", "&nbsp;"*4)
        return self

    def attach(self, attachments: Union[PathLike, Collection[PathLike]]) -> FluentEntity:
        """Attach a file or a collection of files to this message."""
        self.entity.attachments.add([os.fspath(attachment) for attachment in attachments] if isinstance(attachments, Iterable) else os.fspath(attachments))
        return self

    def sign(self, signing: bool = True) -> FluentEntity:
        """Set whether the signature will be appended to the message body prior to sending."""
        self._signing = signing
        return self

    def importance(self, importance_level: str = ImportanceLevel.Normal.value) -> FluentEntity:
        self.entity.importance = importance_level
        return self

    def categories(self, categories: list[str]) -> FluentEntity:
        self.entity.categories = categories
        return self

    def _parse_contacts_to_emails(self, contacts: Union[Union[str, Contact], Collection[Union[str, Contact]]]) -> list[str]:
        from .people import Contact

        if isinstance(contacts, str):
            contact_list = [contacts]
        elif isinstance(contacts, Contact):
            contact_list = [contacts.main_email]
        elif isinstance(contacts, Iterable):
            contact_list = [contact.main_email if isinstance(contact, Contact) else contact for contact in contacts]
        else:
            raise TypeError(f"Expected a single email address or {Contact.__name__}, or an iterable of the above, not '{type(contacts).__name__}'.")

        return contact_list
